/**
 * Copyright (C) 2022 Chris Noxz
 * Author(s): Chris Noxz <chris@noxz.tech>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <errno.h>             // for errno, ENODATA
#include <fcntl.h>             // for open, O_NOFOLLOW, O_RDONLY
#include <limits.h>            // for PATH_MAX
#include <stdarg.h>            // for va_end, va_list, va_start
#include <stdbool.h>           // for true, false, bool
#include <stdio.h>             // for fprintf, stdout, snprintf, sprintf
#include <stdlib.h>            // for free, EXIT_FAILURE, exit, malloc
#include <string.h>            // for strlen, strcmp, memset, strerror
#include <time.h>              // for timespec
#include <unistd.h>            // for close, read
#include <bits/getopt_core.h>  // for getopt, optind
#include <sys/stat.h>          // for stat, fstat, lstat, S_ISREG
#include <sys/xattr.h>         // for fgetxattr, fremovexattr, fsetxattr

#include "sha256.h"            // for SHA256_final, SHA256_init, SHA256_update
#include "acst.h"


static void
bin2hex(unsigned char *bin, size_t len, char *hex)
{
	unsigned int    i,                      // index counter of binary data
	                j;                      // index counter of hex data
	for (i = j = 0; i < len; i++) {
		hex[j++] = hextab[bin[i] >> 4];
		hex[j++] = hextab[bin[i] & 0x0F];
	}
	hex[j] = 0;
}


static int
cscmp(const unsigned char *cs1, const unsigned char *cs2)
{
	const unsigned char \
	               *l,                      // pointer for "left" checksum
	               *r;                      // pointer for "right" checksum
	int             n;                      // countdown from checksum length

	/* iterate through checksums until mismatch or end of checksum */
	for (l = cs1, r = cs2, n = SHA256_BYTES; n && *l == *r; n--, l++, r++);
	return n ? *l - *r : 0; /* return difference between checksums */
}


static void
error(enum Error er, const char *fmt, ...)
{
	va_list         ap;                     // argument list holder

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	switch (er) {
	case ER_NOT_REGULAR:                    cnt.errNotRegular++; break;
	case ER_OPENING:                        cnt.errOpening++; break;
	case ER_XATTR_OPERATION:                cnt.errWritingXattr++; break;
	case ER_GENERIC:                        cnt.errGeneric++; break;
	case ER_FATAL:                          exit(EXIT_FAILURE);
	}
	cnt.errs++;
}


static void
dup_free(dn_t **head_ref)
{
	/* free up memory allocated by nodes, recursively */
	if ((*head_ref)->next)
		dup_free(&((*head_ref)->next));
	free(*head_ref);
}


static void
dup_front_back_split(dn_t *src, dn_t **front_ref, dn_t **back_ref)
{
	dn_t           *f = src->next,          // "fast" pointer
	               *s = src;                // "slow" pointer

	/* advance the "fast" pointer in double the speed as "slow", so that "slow"
	 * will reference the node before the midpoint of the linked list */
	while (f != NULL) {
		f = f->next;
		if (f != NULL)
			s = s->next, f = f->next;
	}

	/* the front will reference the source, while the back will reference the
	 * "slow" pointer or the node before the midpoint of the linked list */
	*front_ref = src;
	*back_ref = s->next;

	/* split the list by breaking the linked list at midpoint */
	s->next = NULL;
}


static void
dup_merge_sort(dn_t **head_ref)
{
	dn_t           *h = *head_ref,          // pointer to reference of head
	               *a = NULL,               // pointer to list a
	               *b = NULL;               // pointer to list b

	/* define base case, start with length 0 or 1 before merge */
	if (h == NULL || h->next == NULL)
		return;

	/* split head into two lists before sorting, until length 0 or 1 */
	dup_front_back_split(h, &a, &b);

	/* repeat merge sort for each list, recursively */
	dup_merge_sort(&a);
	dup_merge_sort(&b);

	/* finally merge list the two lists sorted together into head */
	*head_ref = dup_sorted_merge(a, b);
}


static int
dup_print(dn_t **head_ref)
{
	dn_t           *cur,                    // holder for current node
	               *pre;                    // holder for previous node
	char            hex[CSSZ + 1];          // holder for checksum (as hex)
	int             i, j = 0;               // counters for duplicates

	/* perform a merge sort of linked list before printing to enable detection
	 * of subsequent duplicates */
	dup_merge_sort(head_ref);

	/* iterate through linked list while remembering previous node as reference
	 * for occurrence of duplicates -- if previous and current are identical */
	for (cur = *head_ref, pre = NULL; cur; pre = cur, cur = cur->next) {
		/* check if current and previous match, if so count as duplicates */
		i = (pre && cscmp(cur->cs, pre->cs) == 0) ? i + 1 : 0;
		if (i == 0)             /* ignore non-duplicates */
			continue;
		if (i == 1 && ++j) {    /* include checksum before first filename */
			bin2hex(pre->cs, SHA256_BYTES, hex);
			printf(STR_DPH_FRMT, hex, pre->fn);
		}
		printf(STR_DPN_FRMT, cur->fn);
	}

	/* free up memory when done */
	dup_free(head_ref);

	fprintf(stderr, STR_DPS_FRMT, j);

	return EXIT_SUCCESS;
}


static void
dup_push(dn_t **head_ref, const char *fn, int fd, xa_t *xa)
{
	dn_t           *new = NULL;             // holder for new duplicate node

	/* read checksum from file */
	if (!xa_read(fd, xa) || xa->hex[0] == 0)
		return;

	/* allocate memory for new duplicate node */
	if (!(new = (dn_t*)malloc(sizeof(dn_t))))
		error(ER_FATAL, STR_ERR_OOM);

	/* store checksum and filename together */
	memcpy(new->cs, xa->cs, SHA256_BYTES);
	memcpy(new->fn, fn, MIN(PATH_MAX, strlen(fn)) + 1);

	/* add duplicate node to linked list */
	new->next = *head_ref;
	*head_ref = new;
}


static dn_t*
dup_sorted_merge(dn_t *a, dn_t *b)
{
	dn_t           *r = NULL;               // result holder

	/* define base case, start with length 0 or 1 before merge */
	if (a == NULL || b == NULL)
		return a == NULL ? b : a;

	/* sort by checksum in ascending order (... >= 0 for descending) */
	if (cscmp(a->cs, b->cs) <= 0)
		(r = a)->next = dup_sorted_merge(a->next, b);
	else
		(r = b)->next = dup_sorted_merge(a, b->next);

	/* return sorted merge of a and b, until whole list is sorted */
	return r;
}


static enum FileState
file_state(int fd, xa_t *xa_s, xa_t *xa_a)
{
	if (arg.remove)
		return xa_write(fd, NULL) ? FS_REMOVED : FS_REMOVED_ERROR;

	if (!xa_compute(fd, xa_a))
		return FS_ERROR;
	else if (xa_a->hex[0] == 0)
		return FS_DISRUPTED;

	if (!xa_read(fd, xa_s))
		return FS_MALFORMED;

	if (xa_s->hex[0] == 0)
		return FS_NEW;

	/* compare stored and actual timestamp */
	xa_a->tcmp = (
	      xa_s->s != xa_a->s ? (int)(xa_s->s - xa_a->s)
	    : xa_s->ns != xa_a->ns ? (int)(xa_s->ns - xa_a->ns)
	    : 0
	);

	return (
	      cscmp(xa_s->cs, xa_a->cs) == 0    // if hash is correct
	    ? xa_a->tcmp == 0 ? FS_OK           // ...and ts are matching
	    : FS_TOUCHED                        // ...or if not
	    : xa_a->tcmp < 0 ? FS_OUTDATED      // else if ts is new
	    : xa_a->tcmp > 0 ? FS_BACKDATED     // else if ts is old
	    : FS_CORRUPT                        // else
	);
}


static void
process_argument(const char *fn)
{
	struct stat     st;                     // stat holder

	if (lstat(fn, &st) != 0)
		error(ER_OPENING, STR_ERR_OPNF, fn);
	else if (S_ISREG(st.st_mode))
		process_file(fn);
	/* make sure to only process regular files */
	else
		error(ER_NOT_REGULAR, STR_ERR_REGF, fn);
}


static void
process_file(const char *fn)
{
	int             fd;                     // file holder
	xa_t            xa_s,                   // stored attribute
	                xa_a;                   // actual attribute
	enum FileState  st;                     // file state

	if (!(fd = open(fn, O_RDONLY | O_NOFOLLOW)))
		return error(ER_OPENING, STR_ERR_OPNF, fn);

	if (arg.duplicates)
		dup_push(&gl.dup_head, fn, fd, &xa_s);
	else if ((st = file_state(fd, &xa_s, &xa_a)) == FS_ERROR)
		error(ER_GENERIC, STR_ERR_HASH, fn);
	else if (st == FS_REMOVED_ERROR)
		error(ER_XATTR_OPERATION, STR_ERR_XARM, fn, strerror(errno));
	else {
		report(fn, st);
		if (!arg.quiet && !(st == FS_OK || st == FS_REMOVED || st == FS_DISRUPTED || st == FS_MALFORMED))
			fprintf(stdout, STR_CMP_FRMT
			    , strlen(xa_s.hex) > 0 ? xa_s.hex : zSHA256
			    , xa_s.s
			    , xa_s.ns
			    , strlen(xa_a.hex) > 0 ? xa_a.hex : zSHA256
			    , xa_a.s
			    , xa_a.ns
			);
		if (!(st == FS_OK || st == FS_REMOVED || st == FS_DISRUPTED || st == FS_CORRUPT || st == FS_BACKDATED || st == FS_MALFORMED) && !xa_write(fd, &xa_a))
			error(ER_XATTR_OPERATION, STR_ERR_XAWR, fn);
	}

	close(fd);
}


static void
report(const char *fn, enum FileState st)
{
	switch (st) {
	case FS_OK:
	case FS_REMOVED:                        cnt.ok++; break;
	case FS_DISRUPTED:                      cnt.disrupted++; break;
	case FS_CORRUPT:                        cnt.corrupt++; break;
	case FS_TOUCHED:                        cnt.touched++; break;
	case FS_NEW:                            cnt.new++; break;
	case FS_OUTDATED:                       cnt.outdated++; break;
	case FS_BACKDATED:                      cnt.backdated++; break;
	case FS_MALFORMED:                      cnt.malformed++; break;
	/* don't report errors here (this is only to avoid compiler warnings) */
	case FS_REMOVED_ERROR:
	case FS_ERROR:                          return;
	}
	cnt.total++;

	if (st == FS_CORRUPT || st == FS_BACKDATED || st == FS_MALFORMED) {
		fprintf(stderr, STR_ERR_ABNO, fn);
		fprintf(stdout, STR_OUT_STAT
		    , st == FS_CORRUPT ? STR_OUT_CORR
		    : st == FS_BACKDATED ? STR_OUT_BACK
		    : STR_OUT_MALF, fn);
	}

	if (arg.quiet < 2 && (st == FS_DISRUPTED || st == FS_TOUCHED || st == FS_NEW || st == FS_OUTDATED))
		fprintf(stdout, STR_OUT_STAT
		    , st == FS_DISRUPTED ? STR_OUT_DISR
		    : st == FS_TOUCHED ? STR_OUT_TOUC
		    : st == FS_NEW ? STR_OUT_NEW
		    : STR_OUT_OUTD, fn);

	if (arg.quiet < 1 && (st == FS_OK || st == FS_REMOVED))
		fprintf(stdout, STR_OUT_STAT
		    , st == FS_OK ? STR_OUT_OK
		    : STR_OUT_REMO, fn);
}


static bool
xa_compute(int fd, xa_t *xa)
{
	struct stat     st;                     // file stat holder

	if ((fstat(fd, &st) != 0))
		return false;

	/* retrieve date modified and compute hash of file */
	xa->s = st.st_mtim.tv_sec;
	xa->ns = st.st_mtim.tv_nsec;
	xa_hash(fd, xa->cs);
	bin2hex(xa->cs, SHA256_BYTES, xa->hex);

	/* get file stats after hash computation as the file could have been
	 * changed or deleted during the computation */
	if ((fstat(fd, &st) != 0))
		return false;

	/* check if the file was modified while the hash was computed, and of so
	 * indicate with an empty hash */
	if (!((long unsigned int)st.st_mtim.tv_sec == xa->s && (long unsigned int)st.st_mtim.tv_nsec == xa->ns))
		xa->hex[0] = 0;

	return true;
}


static void
xa_hash(int fd, unsigned char *cs)
{
	struct SHA256_ctx \
	                ctx;                    // SHA256 context
	unsigned char   buf[BUFSZ];             // file content read buffer
	size_t          len;                    // holder for read buffer length

	SHA256_init(&ctx);
	while ((len = read(fd, buf, BUFSZ)))
		SHA256_update(buf, len, &ctx);

	SHA256_final(&ctx, cs);
}


static bool
xa_read(int fd, xa_t *xa)
{
	char           *c;                      // char pointer used in loop (hex)
	unsigned char  *p;                      // char pointer used in loop (bin)
	int             i = 0;                  // index counter to track odd/even

	/* initialize metadata holders to zero, and create a blank xa */
	memset(xa, 0, sizeof(*xa));
	memset(ts, 0, sizeof(*ts));

	/* read checksum, on error indicate as new if missing else fail */
	if (fgetxattr(fd, XATTR_CS, xa->hex, sizeof(xa->hex)) < 0)
		return errno == ENODATA ? true : false;
	else if (strlen(xa->hex) != CSSZ)
		return false;

	/* make sure digits are hex (else fail) and in lower case (else adjust),
	 * and also convert hex digits into binary data for checksum */
	for (c = xa->hex, p = xa->cs; *c; c++) {
		if (*c >= 0x41 && *c <= 0x46)
			*c += 0x20;
		else if ((*c < 0x30 || *c > 0x39) && (*c < 0x61 || *c > 0x66))
			return false;
		if (i++ % 2 == 0)
			*p =  (*c >= 0x30 && *c <= 0x39 ? *c - 0x30
			      :*c >= 0x61 && *c <= 0x66 ? *c - 0x57 : 0) * 16;
		else
			(*p++) += (*c >= 0x30 && *c <= 0x39 ? *c - 0x30
			          :*c >= 0x61 && *c <= 0x66 ? *c - 0x57 : 0);
	}

	/* read timestamp, on error (ENODATA included) fail */
	if (fgetxattr(fd, XATTR_TS, ts, sizeof(ts) - 1) < 0)
		return false;
	else if (sscanf(ts, FRMT_TS, &xa->s, &xa->ns) < 1)
		return false;

	return true;
}


static bool
xa_write(int fd, const xa_t *xa)
{
	if (arg.dryrun)
		return true;

	/* remove extended attributes from file if xa is null */
	if (xa == NULL)
		return (
		    fremovexattr(fd, XATTR_TS) |
		    fremovexattr(fd, XATTR_CS)
		) == 0;

	/* prepare and write extended attributes to file */
	snprintf(ts, sizeof(ts), FRMT_TS, xa->s, xa->ns);
	return (
	    fsetxattr(fd, XATTR_TS, ts, strlen(ts), 0) |
	    fsetxattr(fd, XATTR_CS, &xa->hex, CSSZ, 0)
	) == 0;
}


int
main(int argc, char *argv[])
{
	char            line[PATH_MAX + 1];     // line holder for reading stdin
	int             opt,                    // argument option holder
	                i,                      // index counter for arguments
	                c;                      // holder for getc(stdin)

	/* declare global variables */
	gl.dup_head     = NULL;
	gl.prg          = argv[0];

	while ((opt = getopt(argc, argv, "dhmnqvx")) != -1)
		switch (opt) {
		case 'd':                           arg.duplicates = true; break;
		case 'h':                           USAGE(EXIT_SUCCESS); break;
		case 'm':                           arg.summarize = true; break;
		case 'n':                           arg.dryrun = true; break;
		case 'q':                           arg.quiet++; break;
		case 'v':                           VER(); break;
		case 'x':                           arg.remove = true; break;
		default:                            USAGE(EXIT_FAILURE);
		}

	if ((i = optind) >= argc)
		USAGE(EXIT_FAILURE);

	/* read stdin if the only FILE is -, while the last argument wasn't -- */
	if (argc - i == 1 && argv[i][0] == 0x2d && argv[i - 1][1] != argv[i][0])
		for (i = 0; (c = getc(stdin)) != EOF; i++) {
			if (i > PATH_MAX)
				error(ER_FATAL, STR_ERR_PMAX);
			if ((line[i] = (c != 0x0a) ? c : 0) != 0)
				continue;
			process_argument(line);
			i = -1;
		}
	/* ...else, process each argument as a FILE */
	else
		for (; i < argc; i++)
			process_argument(argv[i]);

	if (arg.duplicates && gl.dup_head)
		return dup_print(&gl.dup_head);

	if (arg.summarize)
		SUMMERIZE();

	if ((cnt.corrupt + cnt.malformed + cnt.backdated) > 0)
		return 5;
	if (cnt.errs == 0 && (cnt.ok + cnt.outdated + cnt.touched + cnt.new) == cnt.total)
		return EXIT_SUCCESS;
	if (cnt.errOpening == cnt.errs)
		return 2;
	if (cnt.errNotRegular == cnt.errs)
		return 3;
	if (cnt.errWritingXattr == cnt.errs)
		return 4;
	return 6;
}
